# mypy: disable-error-code="return"

from os import environ

from pytest import fixture, skip

from ebikeconnect import EBikeConnectClient


@fixture(scope="function")
def client() -> EBikeConnectClient:
    return EBikeConnectClient("cookie")


@fixture(scope="session")
def authenticated_client() -> EBikeConnectClient:
    try:
        client = EBikeConnectClient.login(environ["EBIKE_CONNECT_USERNAME"], environ["EBIKE_CONNECT_PASSWORD"])
        client.PAGE_SIZE = 4  # Set a lower PAGE_SIZE in order to more easily test paging.
        return client
    except KeyError:
        skip("Requires authentication.")



from dataclasses import field
from datetime import datetime, timedelta

from pydantic import ConfigDict, field_validator
from pydantic.dataclasses import dataclass

config = ConfigDict(extra="ignore")


@dataclass(config=config)
class EBikeConnectActivity:
    id: str
    start_time: datetime
    end_time: datetime
    driving_time: timedelta
    type: str
    status: int
    total_distance: float

    @property
    def is_complete(self) -> bool:
        """Returns True if the activity has been fully synchronized."""
        return self.status == 1

    @field_validator("start_time", "end_time", mode="before")
    @classmethod
    def parse_datetime(cls, value):
        if isinstance(value, str):
            value = int(value)
        if isinstance(value, int):
            value = datetime.fromtimestamp(value / 1000).astimezone()
        return value

    @field_validator("driving_time", "operation_time", mode="before")
    @classmethod
    def parse_timedelta(cls, value):
        if isinstance(value, str):
            value = int(value)
        if isinstance(value, int):
            value = timedelta(milliseconds=value)
        return value


@dataclass(config=config)
class EBikeConnectTrip(EBikeConnectActivity):
    ride_ids: list[str]

    @field_validator("type", mode="after")
    @classmethod
    def check_type(cls, value):
        assert value == "NORMAL_TRIP"
        return value


@dataclass(config=config)
class EBikeConnectRide(EBikeConnectActivity):
    title: str
    calories: float
    avg_speed: float
    max_speed: float

    @field_validator("type", mode="after")
    @classmethod
    def check_type(cls, value):
        assert value == "BIKE_RIDE"
        return value


@dataclass(config=config)
class EBikeConnectRideDetails(EBikeConnectActivity):
    operation_time: timedelta
    header_type: str
    calories: float
    avg_speed: float
    max_speed: float
    avg_heart_rate: float
    avg_cadence: float
    avg_altitude: float
    max_heart_rate: int
    max_cadence: int
    max_altitude: float
    cadence: list[list[int | None]] = field(repr=False)
    heart_rate: list[list[int | None]] = field(repr=False)
    speed: list[list[float | None]] = field(repr=False)
    coordinates: list[list[list[float | None]]] = field(repr=False)
    portal_altitudes: list[list[float | None]] = field(repr=False)
    training_effect: int
    training_load_peak: int
    speed_weight: int
    cadence_weight: int
    driver_power_weight: int
    significant: int
    total_driver_power: int
    total_driver_consumption_percentage: float
    total_battery_consumption_percentage: float
    bui_decoded_serial_number: str
    bui_decoded_part_number: str
    drive_unit_decoded_serial_number: str
    drive_unit_decoded_part_number: str
    average_driver_power: float
    power_output: list[list[int]] = field(repr=False)
    significant_assistance_level_percentages: list[dict[str, int | float]] = field(repr=False)
    drive_unit_serial: str
    title: str
    elevation_gain: float | None = None
    elevation_loss: float | None = None

    @field_validator("type", mode="after")
    @classmethod
    def check_type(cls, value):
        assert value == "BIKE_RIDE"
        return value


__all__ = ["EBikeConnectTrip", "EBikeConnectRide", "EBikeConnectRideDetails"]



from itertools import islice

import pytest
from requests.exceptions import HTTPError

from ebikeconnect import EBikeConnectClient


class TestClient:
    def test_invalid_authentication(self):
        with pytest.raises(HTTPError) as exc:
            EBikeConnectClient.login("username", "password")
        assert exc.value.response.status_code == 403
        assert "403 Client Error" in str(exc.value)

    def test_user_agent(self):
        with pytest.raises(HTTPError) as exc:
            EBikeConnectClient.login("username", "password")
        assert exc.value.response.request.headers["User-Agent"].startswith("python-ebikeconnect/")

    def test_get_trip(self, client: EBikeConnectClient):
        with pytest.raises(NotImplementedError):
            client.get_trip("abcd")


class TestAuthenticatedClient:
    def test_cookie(self, authenticated_client: EBikeConnectClient):
        assert len(authenticated_client.cookie) > 30

    def test_find_rides(self, authenticated_client: EBikeConnectClient):
        rides = list(islice(authenticated_client.find_rides(), 6))
        assert len(rides) == 6, "Please ensure you have at least 6 rides."
        ride_ids = {ride.id for ride in rides}
        assert len(ride_ids) == 6  # Verify that paging does not result in duplicates.
        # Now create a new request, skipping the first two results.
        other_rides = list(islice(authenticated_client.find_rides(to_datetime=rides[2].end_time), 4))
        assert len(other_rides) == 4
        other_ride_ids = {ride.id for ride in other_rides}
        assert rides[0].id not in other_ride_ids
        assert rides[1].id not in other_ride_ids
        assert other_ride_ids.issubset(ride_ids)  # Verify paging did not skip a ride.

    def test_find_trips(self, authenticated_client: EBikeConnectClient):
        trips = list(islice(authenticated_client.find_trips(), 6))
        assert len(trips) == 6, "Please ensure you have at least 6 trips."
        trip_ids = {trip.id for trip in trips}
        assert len(trip_ids) == 6  # Verify that paging does not result in duplicates.
        # Now create a new request, skipping the first two results.
        other_trips = list(islice(authenticated_client.find_trips(to_datetime=trips[2].end_time), 4))
        assert len(other_trips) == 4
        other_trip_ids = {ride.id for ride in other_trips}
        assert trips[0].id not in other_trip_ids
        assert trips[1].id not in other_trip_ids
        assert other_trip_ids.issubset(trip_ids)  # Verify paging did not skip a trip.

    def test_get_ride(self, authenticated_client: EBikeConnectClient):
        rides = authenticated_client.find_rides()
        id = next(iter(rides)).id
        ride_details = authenticated_client.get_ride(id)
        assert ride_details.id == id

    def test_get_nonexistent_ride(self, authenticated_client: EBikeConnectClient):
        with pytest.raises(HTTPError) as exc:
            authenticated_client.get_ride("1234")
        assert exc.value.response.status_code == 400



from datetime import datetime
from typing import Any, Iterable

from requests import get, post

from ebikeconnect._version import __version__
from ebikeconnect.models import (
    EBikeConnectRide,
    EBikeConnectRideDetails,
    EBikeConnectTrip,
)


class EBikeConnectClient:
    BASE_URL = "https://www.ebike-connect.com/ebikeconnect/api"
    PAGE_SIZE = 50
    USER_AGENT = "python-ebikeconnect/{}".format(__version__)

    def __init__(self, cookie: str):
        self.cookie = cookie

    @classmethod
    def login(cls, username: str, password: str) -> "EBikeConnectClient":
        response = post(
            cls.BASE_URL + "/portal/login/public",
            headers={"User-Agent": cls.USER_AGENT},
            json=dict(username=username, password=password, rememberme=True),
        )
        response.raise_for_status()
        return cls(response.cookies["REMEMBER"])

    def get_ride(self, id: str) -> EBikeConnectRideDetails:
        data = self._request(url=f"/activities/ride/details/{id}")
        return EBikeConnectRideDetails(**data)

    def get_trip(self, id: str) -> None:
        raise NotImplementedError("The trip details provide no additional information.")

    def find_rides(
        self, from_datetime: datetime | None = None, to_datetime: datetime | None = None
    ) -> Iterable[EBikeConnectRide]:
        response = self._request(
            "/portal/activities/ride/headers",
            parameters={"max": self.PAGE_SIZE, "offset": int((to_datetime or datetime.now()).timestamp() * 1000)},
        )
        last_start_time = None
        for header in response["bike_ride_header_dtos"]:
            assert len(header["header_rides_ids"]) == 1
            assert header["id"] == header["header_rides_ids"][0]
            ride = EBikeConnectRide(**header)
            if from_datetime is not None and ride.start_time < from_datetime:
                return
            yield ride
            last_start_time = ride.start_time
        if last_start_time is not None:
            yield from self.find_rides(from_datetime=from_datetime, to_datetime=last_start_time)

    def find_trips(
        self, from_datetime: datetime | None = None, to_datetime: datetime | None = None
    ) -> Iterable[EBikeConnectTrip]:
        response = self._request(
            "/portal/activities/trip/headers",
            parameters={"max": self.PAGE_SIZE, "offset": int((to_datetime or datetime.now()).timestamp() * 1000)},
        )
        last_start_time = None
        for header in response:
            ride_ids = header["header_rides_ids"]
            trip = EBikeConnectTrip(**header, ride_ids=ride_ids)
            if from_datetime is not None and trip.start_time < from_datetime:
                return
            yield trip
            last_start_time = trip.start_time
        if last_start_time is not None:
            yield from self.find_trips(from_datetime=from_datetime, to_datetime=last_start_time)

    def _request(self, url: str, parameters: dict[str, Any] | None = None) -> Any:
        headers = {"Protect-from": "CSRF", "User-Agent": self.USER_AGENT}
        response = get(
            self.BASE_URL + url,
            headers=headers,
            cookies=dict(REMEMBER=self.cookie),
            params=parameters,
        )
        response.raise_for_status()
        return response.json()



from datetime import datetime

from ebikeconnect.models import EBikeConnectActivity


class TestActivity:
    def test_init(self):
        activity = EBikeConnectActivity(
            **{
                "id": "12345678",
                "start_time": "1723304159000",
                "end_time": "1723304416000",
                "driving_time": "254000",
                "type": "BIKE_RIDE",
                "status": 1,
                "total_distance": 1207.0,
            }
        )
        assert isinstance(activity.start_time, datetime)
